//*****************************************************************************
//   Copyright (c) 1998 Bernard Sufrin, Oxford.
//   http://www.comlab.ox.ac.uk/oucl/people/bernard.sufrin.html
//*****************************************************************************

/**     A bitmap character of size width*height
        is represented (prodigally) by a byte array.
        No more than one pixel is stored per byte
        (indeed there may be fewer -- empty corners
        of the character are cropped, and the values
        of xoffset and yoffset altered accordingly).
        
        <p>

        If the character has been scaled from an unscaled
        character, then the byte at (i,j) is the number of black
        pixels in the unscaled (scale*scale) square centred at
        (scale*i, scale*j).
        
        <p>
        Scaled bytes are used as indexes into linearly
        interpolated colourmaps when characters are shown
        at sizes smaller than their design size. The
        resulting visual appearance is usually more satisfactory
        than the use of threshold based on/off sampled
        bitmaps.
*/

package figue.jape;

public class BitMapChar implements java.io.Serializable {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/jape/BitMapChar.java,v $ $revision$ $Date: 1998/11/02 15:55:21 $ Copyright 1998, Bernard Sufrin, Oxford.";

    public int     width;
    public int     height;
    public int     xoffset;
    public int     yoffset;
    private byte[] bitmap;
   
    private BitMapChar unscaled; // the unscaled model of this character
    private int        scale;    // the scale
   
    public BitMapChar(int width, int height, int xoffset, int yoffset, byte[] bitmap)
    { this.width   = width;
    this.height  = height;
    this.xoffset = xoffset;
    this.yoffset = yoffset;
    this.bitmap  = bitmap;
    this.scale   = 1;
    this.unscaled = this;      // for completeness
    }
   
    public BitMapChar(int width, int height, int xoffset, int yoffset, int scale, byte[] bitmap)
    { this.width   = width;
    this.height  = height;
    this.xoffset = xoffset;
    this.yoffset = yoffset;
    this.bitmap  = bitmap;
    this.scale   = scale;
    this.unscaled = this;      // for completeness
    }
   
    public BitMapChar(int width, int height, int xoffset, int yoffset, int scale, BitMapChar unscaled)
    { this.width    = width;
    this.height   = height;
    this.xoffset  = xoffset;
    this.yoffset  = yoffset;
    this.bitmap   = null;      // calculate this on first demand
    this.scale    = scale;
    this.unscaled = unscaled;
    }
   
    /** 
	Make a scaled, grey, box character which is the same size as this character.
	<p>
	The result is suitable for use as a ``default'' glyph.
    */
   
    public BitMapChar box(int scale)
    { BitMapChar result = new BitMapChar(width, height, xoffset, yoffset, null);
    byte[] bitmap = new byte[width*height];
    byte   tone = (byte) (scale*scale / 4);
    for (int r=0; r<height; r++)
	for (int c=0; c<width; c++) 
	    {    if (height < 5 || width < 5 || r<2 || r>height-3 || c<2 || c>width-3) bitmap[r*width+c] = tone;
	    }
    result.bitmap = bitmap;
    return result;
    }
   
    /** Make a scaled sampled version of this character. 
	<p>
	The <tt>scale</tt> parameter is the size of the edge of
	the sampling square.
	<p>
	The bitmap is constructed on first demand.
    */
   
    public final BitMapChar scale(int scale)
    { if (scale==1) return this;
    return new BitMapChar
	(((xoffset > 0 ? scale - xoffset % scale : - xoffset % scale) + width + scale - 1) / scale,
	 ((yoffset > 0 ? scale - yoffset % scale : - yoffset % scale) + height + scale - 1) / scale,
	 ((xoffset > 0 ? scale - xoffset % scale : - xoffset % scale) + xoffset) / scale,
	 ((yoffset > 0 ? scale - yoffset % scale : - yoffset % scale) + yoffset) / scale,
	 scale,
	 this
	 );
    }
   
    public final BitMapChar quantize(int bits)
    { if (bits==8) return this;
    byte[] original = bitmap();
    byte[] q        = new byte [original.length];
    int    range    = scale*scale; // hue range
    int    thresh   = range / 4;
    for (int i=0; i<original.length; i++)
	{ q[i] = (byte) (original[i] > thresh ? 1 : 0); }
     
    return new BitMapChar(width, height, xoffset, yoffset, 1, q);
    }
   
    public final byte[] bitmap()
    { if (bitmap==null)
	{ 
	    bitmap = new byte[width * height];
	    int xpad = unscaled.xoffset > 0 ? scale - unscaled.xoffset % scale : - unscaled.xoffset % scale;
	    int ypad = unscaled.yoffset > 0 ? scale - unscaled.yoffset % scale : - unscaled.yoffset % scale;
	    for (int row = 0; row < unscaled.height; ++ row) 
		{ int rowIndex = row * unscaled.width;
		int scaledRowIndex = (( row + ypad ) / scale ) * width;
		for (int column = 0; column < unscaled.width; ++ column )
		    if (unscaled.bitmap[rowIndex + column] != 0)
			bitmap[scaledRowIndex + ( column + xpad ) / scale ] += 1;
		}
	    // Release unscaled bitmap reference
	    unscaled = null;
	}
    return bitmap;
    }
}

