//*****************************************************************************
//   Copyright (c) 1998 INRIA.
//   http://www.inria.fr/croap/aioli/modules/figue/web/credits.html#COPYRIGHT
//*****************************************************************************

package figue.resource;

import figue.Constants;
import figue.io.Utility;
import figue.error.ErrorHandler;
import figue.error.ErrorHandlerInterface;
import figue.error.Warning;
import figue.error.ImplementationError;
import figue.error.ShouldNotBeCalledError;
import figue.error.DefaultErrorHandler;
import figue.error.Error;

import java.util.Hashtable;
import java.util.Enumeration;
import java.io.Writer;
import java.io.IOException;

/**
 * [PENDING: update the comments] ResourceFactory;
 * @level advanced
 */

abstract class ResourceFactory extends ErrorHandler implements ResourceFactoryInterface {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/resource/ResourceFactory.java,v $ $revision$ $Date: 1998/11/02 15:55:40 $ Copyright 1998 INRIA.";
  
    protected ResourceFactory(ErrorHandlerInterface anErrorHandler) {
	_errorHandler = anErrorHandler;
	_cacheForStringBuffer = new StringBuffer();
    }
    
    public final Enumeration colors() {
	return getColorRegistry().elements();
    }
    
    public final Enumeration fonts() {
	return getFontRegistry().elements();
    }
    
    public final Enumeration styles() {
	return getStyleRegistry().elements();
    }

    public final Enumeration portableFonts() {
	return getPortableFontRegistry().elements();
    }
    
    protected abstract Hashtable getColorRegistry();
    protected abstract Hashtable getFontRegistry();
    protected abstract Hashtable getStyleRegistry();
    protected abstract Hashtable getPortableFontRegistry();
    
    protected final synchronized String computeKey(int red, int green, int blue) {
	_cacheForStringBuffer.append(red).append('/').append(green).append('/').append(blue);
	final String theResult = _cacheForStringBuffer.toString();
	_cacheForStringBuffer.setLength(0);
	return theResult;
    }
    
    protected final synchronized String computeKey(FontFamily aFamily,
						   FontStyle aStyle,
						   int aPointSize,
						   boolean anUnderlinedFlag) {
	if ( anUnderlinedFlag ) {
	    _cacheForStringBuffer.append(1);
	}
	else {
	    _cacheForStringBuffer.append(0);
	}
	_cacheForStringBuffer.append(aFamily.getFamilyName()).append(aStyle.getName()).append(aPointSize);
	final String theResult = _cacheForStringBuffer.toString();
	_cacheForStringBuffer.setLength(0);
	return theResult;
    }

    protected final synchronized String computeKey(String aName,int aDpi,double aScale) {
	_cacheForStringBuffer.append(aName).append('/').append(aDpi)
	    .append('/').append(aScale);
	final String theResult = _cacheForStringBuffer.toString();
	_cacheForStringBuffer.setLength(0);
	return theResult;
    }
    
    public final boolean isColorDefined(String aMnemonic) {
	return getColorRegistry().containsKey(aMnemonic);
    }
    
    public final ColorInterface getColor(String aMnemonic) 
	throws WrongResourceException {
	final ColorInterface theColor = 
	    (ColorInterface)(getColorRegistry().get(aMnemonic));
	if ( theColor == null ) {
	    throw new WrongResourceException("Unknown color: " + aMnemonic);
	}
	return theColor;
    }
    
    public final ColorInterface doGetColor(String aMnemonic,BasicColor aColor)
	throws WrongResourceException {
	if ( aMnemonic == null ) {
	    return createColor(aMnemonic,aColor);
	}
	else {
	    final Hashtable theColorRegistry = getColorRegistry();
	    if ( theColorRegistry.containsKey(aMnemonic) ) {
		throw new WrongResourceException("Color name already in use: " + 
						 aMnemonic);
	    }
	    final ColorInterface theColor = createColor(aMnemonic,aColor);
	    theColorRegistry.put(aMnemonic,theColor);
	    return theColor;
	}
    }
    
    public final ColorInterface doGetColor(String aMnemonic,
					   int red, int green, int blue) 
	throws WrongResourceException {
	if ( aMnemonic == null ) {
	    return createColor(aMnemonic,red,green,blue);
	}
	else {
	    final Hashtable theColorRegistry = getColorRegistry();
	    if ( theColorRegistry.containsKey(aMnemonic) ) {
		throw new WrongResourceException("Color name already in use: " + 
						 aMnemonic);
	    }
	    final ColorInterface theColor = createColor(aMnemonic,red,green,blue);
	    theColorRegistry.put(aMnemonic,theColor);
	    return theColor;
	}
    }
    
    public final boolean isFontDefined(String aMnemonic) {
	return getFontRegistry().containsKey(aMnemonic);
    }
    
    public final FontInterface getFont(String aMnemonic) 
	throws WrongResourceException {
	final FontInterface theFont = 
	    (FontInterface)(getFontRegistry().get(aMnemonic));
	if ( theFont == null ) {
	    throw new WrongResourceException("Unknown font: " + aMnemonic);
	}
	return theFont;
    }

    public final FontInterface doGetFont(String aMnemonic,
					 FontFamily aFamily,
					 FontStyle aStyle,
					 int aPointSize,
					 boolean anUnderlinedFlag)
	throws WrongResourceException {
	if ( aMnemonic == null ) {
	    return createFont(aMnemonic,aFamily,aStyle,aPointSize,anUnderlinedFlag);
	}
	else {
	    final Hashtable theFontRegistry = getFontRegistry();
	    if ( theFontRegistry.containsKey(aMnemonic) ) {
		throw new WrongResourceException("Font name already in use: " + 
						 aMnemonic);
	    }
	    final FontInterface theFont = createFont(aMnemonic,aFamily,aStyle,aPointSize,anUnderlinedFlag);
	    theFontRegistry.put(aMnemonic,theFont);
	    return theFont;
	}
    }

    public final boolean isStyleDefined(String aMnemonic) {
	return getStyleRegistry().containsKey(aMnemonic);
    }

    public final StyleInterface getStyle(String aMnemonic) 
	throws WrongResourceException {
	final StyleInterface theStyle = 
	    (StyleInterface)(getStyleRegistry().get(aMnemonic));
	if ( theStyle == null ) {
	    throw new WrongResourceException("Unknown style: " + aMnemonic);
	}
	return theStyle;
    }

    public final StyleInterface doGetStyle(String aMnemonic,
					   FontInterface aFont,
					   ColorInterface aColor,
					   ColorInterface aBackground) 
	throws WrongResourceException {
	final Hashtable theStyleRegistry = getStyleRegistry();
	if ( theStyleRegistry.containsKey(aMnemonic) ) {
	    throw new WrongResourceException("Style name already in use: " + 
					     aMnemonic);
	}
	StyleInterface theStyle = new Style(aMnemonic,aFont,aColor,aBackground);
	theStyleRegistry.put(aMnemonic,theStyle);
	return theStyle;
    }

    public final ColorInterface doGetDefaultColor() {
	try {
	    return getColor(Constants.DEFAULT_COLOR);
	}
	catch ( WrongResourceException anException ) {
	    try {
		reportError(new Warning("No default color specified"));
		return doGetColor(Constants.DEFAULT_COLOR,BasicColor.BLACK);
	    }
	    catch ( WrongResourceException anotherException ) {
		throw new ImplementationError("Can't allocate default color");
	    }
	}
    }

    public final ColorInterface doGetDefaultBackground() {
	try {
	    return getColor(Constants.DEFAULT_BACKGROUND);
	}
	catch ( WrongResourceException anException ) {
	    try {
		reportError(new Warning("No default background specified"));
		return doGetColor(Constants.DEFAULT_BACKGROUND,BasicColor.WHITE);
	    }
	    catch ( WrongResourceException anotherException ) {
		throw new ImplementationError("Can't allocate default background");
	    }
	}
    }

    public final FontInterface doGetDefaultFont() {
	try {
	    return getFont(Constants.DEFAULT_FONT);
	}
	catch ( WrongResourceException anException ) {
	    try {
		reportError(new Warning("No default font specified"));
		return doGetFont(Constants.DEFAULT_FONT,FontFamily.HELVETICA,FontStyle.PLAIN,11,false);
	    }
	    catch ( WrongResourceException anotherException ) {
		throw new ImplementationError("Can't allocate default font");
	    }
	}
    }
    
    public final StyleInterface doGetDefaultStyle() {
	if ( getDefaultStyle() == null ) {
	    try {
		setDefaultStyle(getStyle(Constants.DEFAULT_STYLE));
	    }
	    catch ( WrongResourceException anException ) {
		try {
		    reportError(new Warning("No default style specified"));
		    setDefaultStyle(doGetStyle("DEFAULT",doGetDefaultFont(),doGetDefaultColor(),doGetDefaultBackground()));
		}
		catch ( WrongResourceException anotherException ) {
		    throw new ImplementationError("Can't allocate default style");
		}
	    }
	}
	return getDefaultStyle();
    }

    protected abstract StyleInterface getDefaultStyle();
    protected abstract void setDefaultStyle(StyleInterface aStyle);

    public final boolean isPortableFontDefined(String aMnemonic) {
	return getPortableFontRegistry().containsKey(aMnemonic);
    }
    
    public final PortableFontInterface getPortableFont(String aMnemonic) 
	throws WrongResourceException {
	final PortableFontInterface theFont =  
	    (PortableFontInterface)(getPortableFontRegistry().get(aMnemonic));
	if ( theFont == null ) {
	    throw new WrongResourceException("Unknown portable font: " + aMnemonic);
	}
	return theFont;
    }

    public final PortableFontInterface doGetPortableFont(String aMnemonic,
							 String aName,
							 int aDpi,
							 double aScale)
	throws WrongResourceException {
	if ( aMnemonic == null ) {
	    return createPortableFont(aMnemonic,aName,aDpi,aScale);
	}
	else {
	    final Hashtable theFontRegistry = getPortableFontRegistry();
	    if ( theFontRegistry.containsKey(aMnemonic) ) {
		throw new WrongResourceException("Portable font name already in use: " + 
						 aMnemonic);
	    }
	    final PortableFontInterface theFont = createPortableFont(aMnemonic,aName,aDpi,aScale);
	    theFontRegistry.put(aMnemonic,theFont);
	    return theFont;
	}
    }

    public final boolean hasNextErrorHandler() {
	return ( _errorHandler != null );
    }

    public final ErrorHandlerInterface getNextErrorHandler() {
	return _errorHandler;
    }

    public final void manageFatalError(Error anError) {
	throw new ShouldNotBeCalledError();
    }

    public final void manageWarning(Warning aWarning) {
	throw new ShouldNotBeCalledError();
    }

    public final boolean canHandleError() {
	return false;
    }

    public final void writeState(Writer anOutput) throws IOException {
	String theKey;
	boolean theFirstFlag = true;
	boolean theNotEmptyFlag = false;
	theFirstFlag = true;
	theNotEmptyFlag = false;
	Hashtable theRegistry = getFontRegistry();
	Enumeration theIterator = theRegistry.keys();
	while ( theIterator.hasMoreElements() ) {
	    if ( theFirstFlag ) {
		theNotEmptyFlag = true;
		theFirstFlag = false;
		anOutput.write("<Fonts>");
	    }
	    theKey = (String)theIterator.nextElement();
	    ((FontInterface)theRegistry.get(theKey)).writeState(anOutput);
	}
	if ( theNotEmptyFlag ) {
	    anOutput.write("</Fonts>");
	}
	theFirstFlag = true;
	theNotEmptyFlag = false;
	theRegistry = getColorRegistry();
	theIterator = theRegistry.keys();
	while ( theIterator.hasMoreElements() ) {
	    if ( theFirstFlag ) {
		theNotEmptyFlag = true;
		theFirstFlag = false;
		anOutput.write("<Colors>");
	    }
	    theKey = (String)theIterator.nextElement();
	    ((ColorInterface)theRegistry.get(theKey)).writeState(anOutput);
	}
	if ( theNotEmptyFlag ) {
	    anOutput.write("</Colors>");
	}
	theFirstFlag = true;
	theNotEmptyFlag = false;
	theRegistry = getStyleRegistry();
	theIterator = theRegistry.keys();
	while ( theIterator.hasMoreElements() ) {
	    if ( theFirstFlag ) {
		theNotEmptyFlag = true;
		theFirstFlag = false;
		anOutput.write("<Styles>");
	    }
	    theKey = (String)theIterator.nextElement();
	    ((StyleInterface)theRegistry.get(theKey)).writeState(anOutput);
	}
	if ( theNotEmptyFlag ) {
	    anOutput.write("</Styles>");
	}
    }
  
    public ErrorHandlerInterface newErrorHandler() {
	return DefaultErrorHandler.INSTANCE;
    }
    
    public void dispose() {
	_cacheForStringBuffer = null;
	_errorHandler = null;
    }

    private ErrorHandlerInterface _errorHandler;
    private StringBuffer _cacheForStringBuffer;

}

final class Style implements StyleInterface {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/resource/ResourceFactory.java,v $ $revision$ $Date: 1998/11/02 15:55:40 $ Copyright 1998 INRIA.";

    public Style(String aName,
		 FontInterface aFont,
		 ColorInterface aColor,
		 ColorInterface aBackground) {
	_name = aName;
	_font = aFont;
	_color = aColor;
	_background = aBackground;
    }
	
    public final String getName() {
	return _name;
    }
	
    public final FontInterface getFont() {
	return _font;
    }
	
    public final void setFont(FontInterface aFont) {
	_font = aFont;
    }
	
    public final ColorInterface getColor() {
	return _color;
    }
	
    public final void setColor(ColorInterface aColor) {
	_color = aColor;
    }
	
    public final ColorInterface getBackground() {
	return _background;
    }
	
    public final void setBackground(ColorInterface aColor) {
	_background = aColor;
    }
	
    public final String toString() {
	StringBuffer theResult = new StringBuffer("Style<");
	theResult.append(_name).append(">");
	return theResult.toString();
    }

    public final void writeState(Writer anOutput) throws IOException {
	anOutput.write("<Style Name=\"");
	Utility.writeStringAsXmlValue(anOutput,_name);
	anOutput.write("\"");
	if ( _font != null ) {
	    anOutput.write(" Font=\"");
	    Utility.writeStringAsXmlValue(anOutput,_font.getName());
	    anOutput.write("\"");
	}
	if ( _color != null ) {
	    anOutput.write(" Foreground=\"");
	    Utility.writeStringAsXmlValue(anOutput,_color.getName());
	    anOutput.write("\"");
	}
	if ( _background != null ) {
	    anOutput.write(" Background=\"");
	    Utility.writeStringAsXmlValue(anOutput,_background.getName());
	    anOutput.write("\"");
	}
	anOutput.write("/>");
    }
	    
    private String _name;
    private FontInterface _font;
    private ColorInterface _color;
    private ColorInterface _background;
    private String _key;
	
}
